import asyncio
from copy import deepcopy
from typing import Dict, List

from toposort import toposort

from .request import Action
from .internal import recur


class Plan:
    def __init__(self, dependency_graph, execution_plan, object_plan):
        self.dependency_graph = dependency_graph
        self.execution_plan = execution_plan
        self.object_plan = object_plan

    def as_obj(self):
        def _fix(node):
            if isinstance(node, set):
                return list(node)
            return node

        a = {
            "dependencies": recur(_fix, self.dependency_graph),
            "plan": recur(_fix, self.execution_plan),
        }
        return a


class Node:
    def __init__(self, action) -> None:
        self.action = action
        self.event = None

    def export_event(self):
        if self.event is None:
            self.event = asyncio.Event()
        return self.event

    def _signal_event(self):
        if self.event is not None:
            self.event.set()

    async def _run(self, context):
        output = await self.action.run(context)
        if output.failed:
            context.invalidate_dependencies(self.action.name)

        for k, v in output.variables.items():
            context.set_var(k, v)

        return output

    async def run(self, context):
        output = await self._run(context)
        context.set_result(self.action.name, output)
        self._signal_event()
        return output


class Joint:
    def __init__(self, awaits: List[Node], triggers) -> None:
        self.awaits = [n.export_event() for n in awaits]
        self.continuation = triggers

    async def run(self, context):
        await asyncio.gather(*[e.wait() for e in self.awaits])
        return await asyncio.gather(*[c.run(context) for c in self.continuation])


class ObjectPlan:
    def __init__(self, *starting):
        self._joints = []
        self._start = starting

    def joint(self, awaits, triggers):
        self._joints.append(
            Joint(awaits=awaits, triggers=triggers)
        )

    async def run(self, context):
        tasks = [s.run(context) for s in self._start] + [j.run(context) for j in self._joints]
        return await asyncio.gather(*tasks)


def plan_granular(actions: Dict[str, Action]) -> Plan:
    toposort_input = {}
    for k, v in actions.items():
        depends_on = set(el.module for el in v.requires() if not el.is_global)
        toposort_input[k] = depends_on
    # XXX toposort is a underused here
    groups = list(toposort(toposort_input))

    plan = {
        "start": groups[0]
    }
    for k, v in toposort_input.items():
        if v:
            plan.setdefault("joints", []).append({
                "awaits": list(v),
                "triggers": [k]
            })

    return Plan(
        dependency_graph=toposort_input,
        execution_plan=plan,
        object_plan=compile_object_plan(plan, actions),
    )


def populate_plan(plan, actions):
    """Populates plan with nodes"""

    _all = set(plan["start"])
    for joint in plan.get("joints", []):
        _all.update(joint["awaits"])
        _all.update(joint["triggers"])

    nodes = {n: Node(actions[n]) for n in _all}

    def ref_nodes(names):
        return [nodes[n] for n in names]

    plan["start"] = ref_nodes(plan["start"])
    joints = plan.get("joints", [])
    for idx, joint in enumerate(joints):
        joints[idx]["awaits"] = ref_nodes(joints[idx]["awaits"])
        joints[idx]["triggers"] = ref_nodes(joints[idx]["triggers"])
    return plan


def compile_object_plan(plan_definition, actions):
    plan_definition = deepcopy(plan_definition)
    ref_plan = populate_plan(plan_definition, actions)
    ref_joints = ref_plan.get("joints", [])

    plan = ObjectPlan(*ref_plan["start"])
    for ref_joint in ref_joints:
        plan.joint(
            awaits=ref_joint["awaits"],
            triggers=ref_joint["triggers"],
        )
    return plan



from attr import fields_dict
from .exc import ConfigurationError


def assert_fields(cls, obj: dict) -> None:
    cls_fields = fields_dict(cls)
    requested_fields = set(obj)
    diff = requested_fields.difference(cls_fields)
    if diff:
        raise ConfigurationError(
            message="unknown fields",
            details=list(diff),
        )


def recur(callback, node, **kwargs):
    if isinstance(node, (tuple, list)):
        new_node = [recur(callback, el, **kwargs) for el in node]
    elif isinstance(node, dict):
        new_node = {k: recur(callback, v, **kwargs) for k, v in node.items()}
    else:
        new_node = callback(node, **kwargs)
    return new_node



import re
from collections import UserDict

from .exc import VarAlreadyDefinedError, VarNotFountError


class Requirement:
    def __init__(self, module, name):
        self.module = module
        self.name = name

    @property
    def is_global(self):
        return self.module is None

    @property
    def full_name(self):
        if self.module:
            return self.module + "." + self.name
        return self.name


class Expression:
    def __init__(self, captures):
        self.captures = captures

    def resolve(self, vartree):
        items = [c.resolve(vartree) for c in self.captures]
        if len(items) == 1:
            return items[0]
        return "".join((str(el) for el in items))

    def requires(self):
        r = (c.requires() for c in self.captures)
        r = (c for c in r if c is not None)
        return set(r)

    def __str__(self):
        return ", ".join(map(str, self.captures))


class VarTree(UserDict):  # pylint: disable=too-many-ancestors
    def __setitem__(self, key, value):
        if key in self:
            raise VarAlreadyDefinedError(
                message="variable already defined",
                details={"key": key},
            )
        super().__setitem__(key, value)

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            raise VarNotFountError(
                message="variable not found",
                details={"key": key}
            )


class Capture:
    def __init__(self, value, static=True):
        self.value = value

    def __str__(self):
        return "Capture({})".format(self.value)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.value == other.value

    def requires(self):
        return None


class StaticCapture(Capture):
    def resolve(self, _vartree):
        return self.value

    def __str__(self):
        return "StaticCapture({})".format(self.value)


class DynamicCapture(Capture):
    def resolve(self, vartree):
        return vartree[self.value]

    def __str__(self):
        return "DynamicCapture({})".format(self.value)

    def requires(self):
        elements = self.value.split(".", 1)
        if len(elements) == 2:
            return Requirement(module=elements[0], name=elements[1])
        return Requirement(module=None, name=elements[0])


PARAMETER_PATTERN = re.compile(r'(\${([^}]+?)})')


def split(expr):
    found_dynamic = False
    if not isinstance(expr, str):
        return [StaticCapture(expr)], found_dynamic
    start = 0
    end = len(expr)
    out = []
    for el in PARAMETER_PATTERN.finditer(expr):
        ns, ne = el.span()
        if ns > start:
            out.append(StaticCapture(value=expr[start:ns]))
        out.append(DynamicCapture(value=el.group(2)))
        found_dynamic = True
        start = ne

    if start != end:
        out.append(StaticCapture(value=expr[start:end]))
    return out, found_dynamic



from html import escape
from typing import Any
import json

import jmespath

class JmespathResolver:
    def resolve(self, definition: Any, data: dict) -> Any:
        return jmespath.search(definition, data)


class Response:
    def __init__(self, json, failed=False, resolver=None):
        self.json = json
        self.failed = failed
        self._resolver = resolver or JmespathResolver()

    def resolve(self, definition: Any) -> Any:
        return self._resolver.resolve(definition, self.json)


def success(data):
    return Response(data, failed=False)


def failure(data):
    return Response(data, failed=True)


async def from_aiohttp(response, json_only=True, loads=json.loads):
    obj = None
    failed = False

    if response.status > 400:
        failed = True
        obj = {
            "error": escape(await response.text()),
            "code": response.status
        }

    try:
        obj = await response.json(loads=loads)
    except Exception as e:
        if json_only:
            failed = True
        obj = {"text": str(e), "code": response.status}

    return Response(json=obj, failed=failed)





class ArachnoError(Exception):
    def __init__(self, message, details):
        super().__init__(message)
        self.message = message
        self.details = details

    def as_obj(self):
        return {
            "message": self.message,
            "details": self.details,
        }

    def __str__(self):
        return "{}: {}".format(self.message, self.details)


class ValidationError(ArachnoError):
    pass

class VarNotFountError(ArachnoError):
    pass


class VarAlreadyDefinedError(ArachnoError):
    pass


class ConfigurationError(ArachnoError):
    pass



from .exc import ValidationError
from .expr import VarTree
from .planner import plan_granular


class _ExecutionContext:
    def __init__(self, modules, dependency_map=None, variables=None):
        self.modules = modules
        self.dependency_map = dependency_map
        self.result = {}
        self.vartree = VarTree(variables or {})
        self._invalidates = {}

    def set_result(self, k, v):
        self.result[k] = v

    def set_var(self, k, v):
        self.vartree[k] = v

    def action_invalidated(self, action_name):
        return self._invalidates.get(action_name)

    def invalidate_dependencies(self, action_name):
        upstreams = self.dependency_map.get(action_name, [])
        for upstream in upstreams:
            if upstream not in self._invalidates:
                self._invalidates[upstream] = action_name


def validate_variables(request, variables):
    user_vars = set(variables)
    req_vars = request.required_variables

    missing = req_vars.difference(user_vars)
    if missing:
        raise ValidationError(
            message="missing variables",
            details=list(missing),
        )


def validate_modules(request, modules):
    user_mods = set(modules)
    req_operations = request.required_modules
    req_mods = set(e[0] for e in req_operations)

    missing = req_mods.difference(user_mods)
    if missing:
        raise ValidationError(
            message="unknown modules",
            details=list(missing),
        )

    unknown_operations = set()
    for mod, operation_name in req_operations:
        module = modules[mod]
        operation = module.get_operation(operation_name)
        if not operation:
            unknown_operations.add(f"{mod}.{operation_name}")

    if unknown_operations:
        raise ValidationError(
            message="unknown operations",
            details=list(unknown_operations),
        )



async def execute(request, modules, variables=None, explain=False):
    variables = variables or {}
    validate_variables(request, variables)
    validate_modules(request, modules)

    plan = plan_granular(request.actions)
    context = _ExecutionContext(
        modules=modules,
        variables=variables,
        dependency_map=build_dependency_map(plan.dependency_graph)
    )

    await plan.object_plan.run(context)

    out = {
        "result": context.result,
    }

    if explain:
        out["plan"] = plan.as_obj()

    return out


def build_dependency_map(dependency_graph):
    dm = {}
    for k, vs in dependency_graph.items():
        for v in vs:
            dm.setdefault(v, set()).add(k)
    return dm




from .request import Request, Action, Options, Duration
from .executor import execute

class BaseModule:
    async def dispatch(self, operation_name, **kwargs):
        operation = self.get_operation(operation_name)
        kwargs = self.prepare_arguments(operation_name, **kwargs)
        return await operation(**kwargs)

    def prepare_arguments(self, operation_name, **kwargs):
        return kwargs

    def get_operation(self, operation_name):
        return getattr(self, operation_name, None)



import asyncio
import time
from copy import copy, deepcopy
from datetime import timedelta
from typing import Any, Dict

import attr.validators as attrv
from attr import attrib, attrs
from cached_property import cached_property
from typing_extensions import Protocol

from .duration import from_str as duration_from_str
from .exc import ConfigurationError, ValidationError
from .expr import Expression
from .expr import split as split_expression
from .internal import assert_fields, recur


class Resolver(Protocol):
    def resolve(self, definition: Any) -> Any:
        pass


@attrs
class Duration:
    label = attrib(validator=attrv.instance_of(str))
    value = attrib(validator=attrv.optional(attrv.instance_of(timedelta)))

    @classmethod
    def from_str(cls, label):
        if not label:
            return cls(label="none", value=None)
        return cls(
            label=label,
            value=duration_from_str(label)
        )

    def seconds(self):
        if self.value is None:
            return None
        return self.value.total_seconds()  # pylint: disable=no-member


@attrs
class Options:
    timeout = attrib(validator=attrv.instance_of(Duration), convert=Duration.from_str, default="")
    suppress = attrib(validator=attrv.instance_of(bool), default=False)

    @classmethod
    def from_obj(cls, obj):
        assert_fields(cls, obj)
        return cls(**obj)


@attrs
class Request:
    actions: Dict[str, "Action"] = attrib(validator=attrv.instance_of(dict))

    def __attrs_post_init__(self):
        self._validate()

    @classmethod
    def from_obj(cls, obj):
        assert_fields(cls, obj)
        obj = deepcopy(obj)
        obj["actions"] = {
            n: Action.from_obj(a, name=n)
            for n, a in obj["actions"].items()
        }
        return cls(**obj)

    @property
    def required_modules(self):
        return copy(self._required_modules)

    @property
    def required_variables(self):
        return copy(self._required_variables)

    @cached_property
    def _required_variables(self):
        """Returns variables needed to execute this request"""

        out = set()
        for action in self.actions.values():  # pylint: disable=no-member
            requirements = action.requires()
            for requirement in requirements:
                if requirement.is_global:
                    out.add(requirement.full_name)
        return out

    @cached_property
    def _required_modules(self):
        out = set()
        for action in self.actions.values():  # pylint: disable=no-member
            out.add((action.module, action.operation))
        return out

    def _validate(self):
        self._validate_dependencies()

    def _validate_dependencies(self):
        definitions = self.required_variables
        args = set()
        requesters = {}

        for action_name, action in self.actions.items():  # pylint: disable=no-member
            for definition in action.defines:
                global_key = action_name + "." + definition
                definitions.add(global_key)

            requirements = action.requires()
            for requirement in requirements:
                full_name = requirement.full_name
                if requirement.module == action_name:
                    raise ValidationError(
                        message="reference to self-defined arg",
                        details={"arg": full_name}
                    )

                requesters.setdefault(full_name, []).append(action_name)
                args.add(full_name)

        not_defined = args.difference(definitions)
        if not_defined:
            tmpl = "required by {} but not defined"
            errs = {k: tmpl.format(
                ",".join(requesters[k])) for k in not_defined}
            raise ValidationError(
                message="undefined arguments",
                details=errs
            )


@attrs
class Action:
    module = attrib(validator=attrv.instance_of(str))
    operation = attrib(validator=attrv.instance_of(str))
    args = attrib(validator=attrv.instance_of(dict), factory=dict)
    defines = attrib(validator=attrv.instance_of(dict), factory=dict)
    options = attrib(validator=attrv.instance_of(Options), factory=Options)
    name = attrib(validator=attrv.optional(attrv.instance_of(str)), default=None)

    @classmethod
    def from_obj(cls, obj, name=None):
        assert_fields(cls, obj)
        obj["module"], obj["operation"] = _parse_operation(obj)
        obj["args"] = wrap_arguments(obj.get("args", {}))
        if name is not None:
            obj["name"] = name
        if obj.get("options"):
            obj["options"] = Options.from_obj(obj["options"])
        return cls(**obj)

    def requires(self):
        out = []

        def _collect(node):
            if isinstance(node, Expression):
                out.extend(node.requires())
        recur(_collect, self.args)
        return set(out)

    async def _run(self, context, loop=None):
        args = substitute_args(self.args, context.vartree)
        module = context.modules[self.module]
        coro = module.dispatch(self.operation, **args)
        start = time.monotonic()
        result = await asyncio.wait_for(coro, timeout=self.options.timeout.seconds(), loop=loop)  # pylint: disable=no-member
        return result, time.monotonic() - start

    async def run(self, context, loop=None):
        reason = context.action_invalidated(self.name)
        if reason is not None:
            return Output(
                failed=True,
                failure_details={
                    "type": "invalidated",
                    "value": reason,
                },
            )

        try:
            result, took = await self._run(context, loop=loop)
        except asyncio.TimeoutError:
            return Output(
                failed=True,
                failure_details={
                    "type": "timeout",
                    "value": self.options.timeout.label,  # pylint: disable=no-member
                },
            )

        if result.failed:
            return Output(
                failed=True,
                took=took,
                failure_details={
                    "type": "module_error",
                    "value": result.json,
                },
            )

        return Output(
            variables=self._resolve_defines(result) if not result.failed else {},
            failed=False,
            data=result.json if not self.options.suppress else None,  # pylint: disable=no-member
            took=took,
        )

    def _resolve_defines(self, resolver: Resolver) -> Dict[str, Any]:
        return {
            "{}.{}".format(self.name, key): resolver.resolve(definition)
            for key, definition in self.defines.items()  # pylint: disable=no-member
        }


@attrs
class Output:
    failed = attrib(validator=attrv.instance_of(bool))
    failure_details = attrib(validator=attrv.instance_of(dict), factory=dict)
    variables = attrib(validator=attrv.instance_of(dict), factory=dict)
    data = attrib(factory=dict)
    took = attrib(default=None)


def _parse_operation(obj):
    module = obj.get("module")
    operation = obj.get("operation")

    if module and operation:
        return module, operation

    if operation and "." in operation:
        return operation.split(".", 1)

    raise ConfigurationError(
        message="invalid operation",
        details={
            "module": module,
            "operation": operation
        })


def wrap_arguments(args):
    return recur(_wrap_expression, args)


def _wrap_expression(value):
    if not isinstance(value, str):
        return value

    expr, is_dynamic = split_expression(value)
    if is_dynamic:
        return Expression(expr)
    return value


def substitute_args(args, vartree):
    def _resolve(node, vartree=None):
        if isinstance(node, Expression):
            return node.resolve(vartree)
        return node
    return recur(_resolve, args, vartree=vartree)



import os
from setuptools import setup, find_packages


REQUIREMENTS_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), 'requirements.txt')
)


def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith('#')]


setup(
    name='arachno',
    version='0.3.0',
    url='https://github.com/palestamp/arachno',
    author='Stas Kazhavets',
    author_email='stas.kozhevets@gmail.com',
    description='DSL for coroutine orchestration',
    packages=find_packages(exclude=['tests']),
    install_requires=parse_requirements(REQUIREMENTS_PATH),
    long_description=open('README.rst').read(),
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
    ]
)
